defmodule Jason.DataStructures do
  @moduledoc """
  Various tools for processing JSON documents.

  - `test01/1` -- A simple test of `iter_apply/4`.
  - `test02/0` -- A simple test of `collect/2`.
  - `collect/2` -- Collect all nodes from a JSON tree that satisfy a predicate function.
  - `iter_apply/4` -- Iterate over all nodes in a JSON document.
    If `predicate` is true, apply `function`.
  - `parsejsondoc/1` -- Parse a JSON document and return the tree/map.
  - `walk/2` -- Walk a JSON map.  Print out each nested item.
  - `iter_map/4` -- Iterate over all nodes in a JSON document.
    Apply functions to produce (map to) a new JSON doc.

  """

  @doc """
  Parse a JSON document and return the tree/map.

  ## Examples

      iex> d1 = Test21.parsejsondoc("Data/array.data01.json")

  """
  @spec parsejsondoc(String.t()) :: map()
  def parsejsondoc(path) do
    #schema_path = Path.join(__DIR__, path)
    content = File.read!(path)
    data = Jason.decode!(content)
    data
  end

  @doc """
  Walk a JSON map.  Print out each nested item.

  ## Examples

      iex> d1 = Test21.parsejsondoc("Data/array.data01.json")
      iex> Test21.walk(d1)

  """
  @spec walk(any(), String.t()) :: atom()
  def walk(data, indent \\ "") do
    cond do
      is_list(data) ->
        Enum.each(data, fn item ->
          walk(item, indent)
        end)
        :list
      is_map(data) ->
        Enum.each(Map.keys(data), fn key ->
          #walk({key, Map.get(data, key)}, indent)
          item = Map.get(data, key)
          IO.puts(~s/#{indent}key: "#{key}"/)
          walk(item, indent <> "    ")
        end)
        :map
      is_binary(data) ->
        IO.puts(~s/#{indent}string: "#{data}"/)
        :string
      is_number(data) ->
        IO.puts(~s/#{indent}number: #{data}/)
        :number
      is_atom(data) ->
        IO.puts(~s/#{indent}atom: #{data}/)
        :atom
      true ->
        IO.inspect(data, label: "unknown data type")
        :error
    end
  end

  @doc """
  Iterate over all nodes in a JSON document.
  If `predicate` is true, apply `function`.

  ## Examples

      iex> data1 = Test21.parsejsondoc("path/to/mydata.json")
      iex> pred1 = fn x -> is_binary(x) end
      iex> func1 = fn x -> IO.puts(~s/string: "\#{x}"/) end
      iex> Test21.iter_apply(data1, pred1, func1)
      string: "apple"
      string: "orange"
      string: "pear"
      string: "potato"
      string: "broccoli"

  """
  @spec iter_apply(any(), (any() -> boolean()), (any(), any() -> any()), any()) :: any()
  def iter_apply(data, predicate, function, args \\ nil) do
    if predicate.(data) do
      function.(data, args)
    end
    cond do
      is_list(data) ->
        Enum.each(data, fn item ->
          iter_apply(item, predicate, function, args)
        end)
        :list
      is_map(data) ->
        Enum.each(Map.keys(data), fn key ->
          item = Map.get(data, key)
          iter_apply(item, predicate, function, args)
        end)
        :map
      is_binary(data) ->
        :string
      is_number(data) ->
        :number
      is_atom(data) ->
        :atom
      true ->
        IO.inspect(data, label: "unknown data type")
        :error
    end
    :ok
  end

  @doc """
  Iterate over all nodes in a JSON document.
  Apply functions to produce (map to) a new JSON doc.

  ## Options

  * `map_function` -- A function of 1 argument to be applied to maps.
  * `list_function` -- A function of 1 argument to be applied to lists.
  * `string_function` -- A function of 1 argument to be applied to strings.
  * `number_function` -- A function of 1 argument to be applied to numbers.

  ## Examples

      def demo_map01(path) do
        string_func = fn x -> x <> x end
        map_func = fn old_map ->
          new_map = Enum.reduce(Map.keys(old_map), Map.new(), fn old_key, acc ->
            new_key = String.upcase(old_key)
            acc1 = Map.put(acc, new_key, Map.get(old_map, old_key))
            acc1
          end)
          new_map
        end
        list_func = fn x -> Enum.map(x, fn y ->
          if is_binary(y) do
            String.reverse(y)
          else
            y
          end
        end)
        end
        data = parsejsondoc(path)
        iter_map(data,
          map_function: map_func,
          string_function: string_func,
          list_function: list_func)
      end

  """
  @spec iter_map(map(), keyword()) :: map()
  def iter_map(data, options \\ []) do
    map_function = Keyword.get(options, :map_function)
    list_function = Keyword.get(options, :list_function)
    string_function = Keyword.get(options, :string_function)
    number_function = Keyword.get(options, :number_function)
    atom_function = Keyword.get(options, :number_function)
    iter_map_aux(data, map_function, list_function, string_function, number_function, atom_function)
  end

  @spec iter_map_aux(any(),
    (any() -> any()),
    (any() -> any()),
    (any() -> any()),
    (any() -> any()),
    (any() -> any())
  ) :: any()
  defp iter_map_aux(data, map_function, list_function, string_function, number_function, atom_function) do
    result = cond do
      is_map(data) ->
        data1 = if is_nil(map_function) do
          data
        else
          data2 = map_function.(data)
          data2
        end
        map1 = Map.new()
        Enum.reduce(Map.keys(data1), map1, fn key, acc ->
          value = Map.get(data1, key)
          value1 = iter_map_aux(value, map_function, list_function, string_function, number_function, atom_function)
          acc1 = Map.put(acc, key, value1)
          acc1
        end)
      is_list(data) ->
        data1 = if is_nil(list_function) do
          data
        else
          list_function.(data)
        end
        Enum.map(data1, fn item ->
          iter_map_aux(item, map_function, list_function, string_function, number_function, atom_function)
        end)
      is_binary(data) ->
        if is_nil(string_function) do
          data
        else
          string_function.(data)
        end
      is_number(data) ->
        if is_nil(number_function) do
          data
        else
          number_function.(data)
        end
      is_atom(data) ->
        if is_nil(atom_function) do
          data
        else
          atom_function.(data)
        end
      true ->
        IO.inspect(data, label: "unknown data type")
        :error
    end
    result
  end

  @doc """
  A test for `iter_map`.
  """
  @spec demo_map01(Path.t()) :: any()
  def demo_map01(path) do
    string_func = fn x -> x <> x end
    map_func = fn old_map ->
      new_map = Enum.reduce(Map.keys(old_map), Map.new(), fn old_key, acc ->
        new_key = String.upcase(old_key)
        acc1 = Map.put(acc, new_key, Map.get(old_map, old_key))
        acc1
      end)
      new_map
    end
    list_func = fn x -> Enum.map(x, fn y ->
      if is_binary(y) do
        String.reverse(y)
      else
        y
      end
    end)
    end
    data = parsejsondoc(path)
    iter_map(data,
      map_function: map_func,
      string_function: string_func,
      list_function: list_func)
  end

  @doc """
  Another test for `iter_map`.
  """
  @spec demo_map02(any()) :: any()
  def demo_map02(data) do
    string_func = fn x -> x <> x end
    map_func = fn old_map ->
      new_map = Enum.reduce(Map.keys(old_map), Map.new(), fn old_key, acc ->
        new_key = String.upcase(old_key)
        acc1 = Map.put(acc, new_key, Map.get(old_map, old_key))
        acc1
      end)
      new_map
    end
    list_func = fn x -> Enum.map(x, fn y ->
      if is_binary(y) do
        String.reverse(y)
      else
        y
      end
    end)
    end
    iter_map(data,
      map_function: map_func,
      string_function: string_func,
      list_function: list_func)
  end

  @doc """
  Collect all nodes from a JSON tree that satisfy a predicate function.

  ## Args
  - `data` -- the JSON tree to be searched; an Elixir map.
  - `predicate` -- A function of one argument; should return a boolean.

  ## Examples

      iex> data1 = Test21.parsejsondoc("Data/array.data01.json")
      iex> pred1 = fn x -> x == "potato" end
      iex> result = Test21.collect(data1, pred1)

  """
  @spec collect(map(), (any() -> boolean())) :: [any()]
  def collect(data, predicate) do
    cond do
      is_list(data) ->
        collection = Enum.reduce(data, [], fn item, acc ->
          items = collect(item, predicate)
          if length(items) > 0 do
            items ++ acc
          else
            acc
          end
        end)
        if predicate.(data) do
          [data | collection]
        else
          collection
        end
      is_map(data) ->
        collection = Enum.reduce(Map.keys(data), [], fn key, acc ->
          item = Map.get(data, key)
          items = collect(item, predicate)
          if length(items) > 0 do
            items ++ acc
          else
            acc
          end
        end)
        if predicate.(data) do
          [data | collection]
        else
          collection
        end
      is_binary(data) or is_number(data) or is_atom(data) ->
        if predicate.(data) do
          [data]
        else
          []
        end
      true ->
        IO.inspect(data, label: "unknown data type")
        :error
    end
  end

  @doc """
  A simple test of `iter_apply/4`.

  ## Examples

      iex> Test21.test01("Data/array.data01.json")

  """
  @spec test01(String.t()) :: :ok
  def test01(path) do
    data1 = Test21.parsejsondoc(path)
    pred1 = fn x -> is_binary(x) end
    func1 = fn item, {label1, label2} -> IO.puts(~s/#{label1}"#{item}"#{label2}/) end
    iter_apply(data1, pred1, func1, {"the item ==", "=="})
    :ok
  end

  @doc """
  A simple test of `collect/2`.

  ## Examples

      iex> Test21.test02()

  """
  @spec test02() :: :ok
  def test02() do
    data1 = Test21.parsejsondoc("Data/array.data01.json")
    pred1 = fn x -> x == "potato" end
    pred2 = fn x -> x == "potato" or x == "broccoli" end
    pred3 = fn x -> is_map(x) and Map.get(x, "veggieLike") == true end
    result1 = Test21.collect(data1, pred1)
    result2 = Test21.collect(data1, pred2)
    result3 = Test21.collect(data1, pred3)
    IO.inspect(result1, label: "result1")
    IO.inspect(result2, label: "result2")
    IO.inspect(result3, label: "result3")
    :ok
  end

end
